
/*
    VPPP virtual PPP connection over TCP.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max_mk@yahoo.com>

    Idea is based on Sam Lantinga's PPPTCP package, but code was 
    completely written from the scratch.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/

/*
 * Version: 2.1 03/06/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */ 

#include <fcntl.h>
#include <stdlib.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <strings.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>

#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sched.h>

#include "vppp.h"
#include "link_fd.h"

#include <zlib.h>

extern int errno;

/* Link with shaping */

int link_fd_err = 0;
int link_fd_len = 0; 
int link_fd_fd1 = 0;
int link_fd_fd2 = 0;

fd_set link_fd_set;
struct timeval link_fd_tv={0,0};
     
char * link_fd_buf;

void link_fd_copy(int sig){
     register int len;

     FD_ZERO(&link_fd_set);
     FD_SET(link_fd_fd2, &link_fd_set);

     /* 
      * Check for available data, or we will sleep in the read call 
      */
     len=select(link_fd_fd2+1, &link_fd_set, NULL, NULL, &link_fd_tv);
  
     if(len == -1){
	link_fd_err = 1;
	return;
     }
  
     if(!len)
	return;

     if((len=read(link_fd_fd2, link_fd_buf, link_fd_len)) <= 0)
 	link_fd_err = 1;
     else 
	if( write_n(link_fd_fd1, link_fd_buf, len) < 0 )
	   link_fd_err = 1;
}

int link_fd_shp(int fd1, int fd2, int speed){
     struct itimerval itm;
     struct sigaction sa;
     char *buf;
     int old_prio;

     if(speed < 64)
	  return 1;

     /*
      * Calculate how much we will copy each timer tick. 
      * It is very primitive shaping, but shaping results 
      * is prety accurate.
      */
 
     link_fd_len=LINKFD_MAGIC * (speed / 64);

     if( !(link_fd_buf=malloc(link_fd_len)) )
         return 1; 

     if( !(buf=malloc(LINKFD_BUF_SIZE)) )
        return 1; 

     link_fd_fd1 = fd1;
     link_fd_fd2 = fd2;
     link_fd_err = 0;

     /* Set up the SIGALRM handler */
     memset(&sa,0,sizeof(sa));    
     sa.sa_handler=link_fd_copy;
     sa.sa_flags |= SA_RESTART;
     sigemptyset(&sa.sa_mask);
     sigfillset(&sa.sa_mask);
     sigaction(SIGALRM,&sa,NULL);    

     /* Increase our priority and lock buf memmory */
     mlock(link_fd_buf,link_fd_len*2);
     mlock(buf,LINKFD_BUF_SIZE);
     old_prio=getpriority(PRIO_PROCESS,0);
     setpriority(PRIO_PROCESS,0,LINKFD_PRIO);

     /* Start the timer */
     memset(&itm,0,sizeof(itm));
     itm.it_interval.tv_usec = itm.it_value.tv_usec = LINKFD_FREQ;
     setitimer(ITIMER_REAL,&itm,NULL);
     
     while(!link_fd_err){
        register int len;

	if( (len=read(link_fd_fd1, buf, LINKFD_BUF_SIZE)) <= 0 )
	      break; 
	if( write_n(link_fd_fd2, buf, len) < 0 )
   	   break; 
     }

     signal(SIGALRM,SIG_DFL);  
 
     setpriority(PRIO_PROCESS,0,old_prio);
     munlock(link_fd_buf,link_fd_len*2);
     munlock(buf,LINKFD_BUF_SIZE);

     memset(&itm,0,sizeof(itm));
     setitimer(ITIMER_REAL,&itm,NULL);

     free(link_fd_buf);
	
     return 0;
}

/* Fastest standard link */

int link_fd_std(int fd1,int fd2){
     int maxfd, old_prio;
     fd_set fdset;
     char *buf;

     if( !(buf=malloc(LINKFD_BUF_SIZE)) )
        return 1; 

     maxfd = (fd1 > fd2 ? fd1 : fd2) + 1;

     /* Increase our priority and lock buf memmory */
     mlock(buf,LINKFD_BUF_SIZE);
     old_prio=getpriority(PRIO_PROCESS,0);
     setpriority(PRIO_PROCESS,0,LINKFD_PRIO);

     while(1){
        register int len;

        /* Wait for data */
        FD_ZERO(&fdset);
	FD_SET(fd1, &fdset);
	FD_SET(fd2, &fdset);
	if(select(maxfd, &fdset, NULL, NULL, NULL) < 0 )
	   break;

	if( FD_ISSET(fd1, &fdset) ){
	   if( (len=read(fd1, buf, LINKFD_BUF_SIZE)) <= 0 )
		break;  
	   if( write_n(fd2, buf, len) < 0 )
		break; 
	}
	if( FD_ISSET(fd2, &fdset) ){
	   if( (len=read(fd2, buf, LINKFD_BUF_SIZE)) <= 0 )
		break;
	   if( write_n(fd1, buf, len) < 0 )
		break;
	}
     }
     setpriority(PRIO_PROCESS,0,old_prio);
     munlock(buf,LINKFD_BUF_SIZE);

     free(buf);

     return 0;
}

/* Link with compression */

int link_fd_cmp(int fd1,int fd2){
     int maxfd, old_prio, err;
     char *in_buf, *out_buf;
     z_stream zi,zd; 
     fd_set fdset;

     maxfd = (fd1 > fd2 ? fd1 : fd2) + 1;

     zd.zalloc = (alloc_func)0;
     zd.zfree  = (free_func)0;
     zd.opaque = (voidpf)0;
     zi.zalloc = (alloc_func)0;
     zi.zfree  = (free_func)0;
     zi.opaque = (voidpf)0;
    
     if( deflateInit(&zd, Z_BEST_SPEED ) != Z_OK ){
	syslog(LOG_ERR,"Can't initialize compressor");
	return 1;
     }	
     if( inflateInit(&zi) != Z_OK ){
	syslog(LOG_ERR,"Can't initialize decompressor");
	return 1;
     }	
     if( !(in_buf = malloc(LINKFD_BUF_SIZE)) ){
	syslog(LOG_ERR,"Can't allocate buffer\n");
	return 1;
     }	
     if( !(out_buf = malloc(LINKFD_BUF_SIZE*3)) ){
	syslog(LOG_ERR,"Can't allocate buffer\n");
	return 1;
     }	

     /* Increase our priority and lock buf memmory */
     mlock(in_buf,LINKFD_BUF_SIZE);
     mlock(out_buf,LINKFD_BUF_SIZE*3);
     old_prio=getpriority(PRIO_PROCESS,0);
     setpriority(PRIO_PROCESS,0,LINKFD_PRIO);

     while(1){
        register int len;

        /* Wait for data */
        FD_ZERO(&fdset);
	FD_SET(fd1, &fdset);
	FD_SET(fd2, &fdset);
	if(select(maxfd, &fdset, NULL, NULL, NULL) < 0 )
	   break;

	if( FD_ISSET(fd1, &fdset) ){
	   if( (len=read(fd1, in_buf, LINKFD_BUF_SIZE)) <= 0 )
		break;
	   zi.next_in = in_buf;
	   zi.avail_in = len;
	   while(zi.avail_in){
		zi.next_out = out_buf;
		zi.avail_out = LINKFD_BUF_SIZE*3;
		if((err=inflate(&zi, Z_SYNC_FLUSH)) != Z_OK){
		   syslog(LOG_ERR,"Inflate error %d\n",err);
		   break;
	        }
		if(write_n(fd2, out_buf, LINKFD_BUF_SIZE*3 - zi.avail_out) < 0)
	           goto stop; 
	   }
	}

	if( FD_ISSET(fd2, &fdset) ){
	   if( (len=read(fd2, in_buf, LINKFD_BUF_SIZE)) <= 0 )
		break; 
	   zd.next_in = in_buf;
	   zd.avail_in = len;
	   while(zd.avail_in){
		zd.next_out = out_buf;
		zd.avail_out = LINKFD_BUF_SIZE*3;
		if( (err=deflate(&zd, Z_SYNC_FLUSH)) != Z_OK ){
		   syslog(LOG_ERR,"Deflate error %d\n",err);
		   break;
		}
		if(write_n(fd1, out_buf, LINKFD_BUF_SIZE*3 - zd.avail_out) < 0)
	           goto stop; 
	   }
	}

     }

stop:
     setpriority(PRIO_PROCESS,0,old_prio);
     munlock(in_buf,LINKFD_BUF_SIZE);
     munlock(out_buf,LINKFD_BUF_SIZE*3);

     deflateEnd(&zd);
     inflateEnd(&zi);

     free(in_buf);
     free(out_buf);

     return 0;
}
